package ru.dbotthepony.mc.prng.mixin;

import net.minecraft.util.RandomSource;
import net.minecraft.world.level.levelgen.LegacyRandomSource;
import net.minecraft.world.level.levelgen.WorldgenRandom;
import net.minecraft.world.level.levelgen.XoroshiroRandomSource;
import org.objectweb.asm.Opcodes;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Overwrite;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.asm.mixin.injection.At;
import org.spongepowered.asm.mixin.injection.Redirect;
import ru.dbotthepony.mc.prng.GJRAND64RandomSource;

@Mixin(WorldgenRandom.class)
public abstract class WorldgenRandomMixin {
	@Shadow
	private RandomSource randomSource;

	// This is a mixed bag os a solution.
	// On one hand, we cover all cases where WorldgenRandom getting called within vanilla code,
	// on other hand it also detours and redirects calls done by mods, which *may* delibarately use LCG or Xoroshiro128++
	// But we also need to consider that mods will likely just copy/do what vanilla code does, and hence
	// they expect code to follow same behavior as vanilla, so in such case us redirecting PRNG is completely desirable
	@Redirect(
		method = "<init>(Lnet/minecraft/util/RandomSource;)V",
		at = @At(
			value = "FIELD",
			target = "randomSource:Lnet/minecraft/util/RandomSource;",
			opcode = Opcodes.PUTFIELD
		)
	)
	public void constructorMix(WorldgenRandom self, RandomSource randomSource) {
		if (randomSource instanceof LegacyRandomSource rng) {
			this.randomSource = new GJRAND64RandomSource(rng.seed.get());
		} else if (randomSource instanceof XoroshiroRandomSource rng) {
			this.randomSource = new GJRAND64RandomSource(rng.randomNumberGenerator.seedHi, rng.randomNumberGenerator.seedLo);
		} else {
			this.randomSource = randomSource;
		}
	}

	@Overwrite
	public static RandomSource seedSlimeChunk(int chunkX, int chunkZ, long levelSeed, long salt) {
		return new LegacyRandomSource(
			levelSeed
				+ (long)(chunkX * chunkX * 4987142)
				+ (long)(chunkX * 5947611)
				+ (long)(chunkZ * chunkZ) * 4392871L
				+ (long)(chunkZ * 389711)
				^ salt
		);
	}
}
